from timelength.constants import CONNECTORS, ABBREVIATIONS

def isfloat(num: str):
    try:
        float(num)
        return True
    except ValueError:
        return False

def check_alphanumeric(char: str):
    if isfloat(char):
        return 1
    elif char.isalpha():
        return 2
    else:
        return None

def check_valid_values(buffer: str):
    if isfloat(buffer) or buffer in CONNECTORS or buffer in ABBREVIATIONS:
        return True
    else:
        return False


class Millisecond(object):
    def __init__(self):
        self.scale = 0.001
        self.terms = ["ms", "millisecond", "Millisecond", "MILLISECOND", "milliseconds", "Milliseconds", "MILLISECONDS", "mil", "Mil", "MIL", "mils", "Mils", "MILS", "milsec", "Milsec", "MILSEC", "milsecs", "Milsecs", "MILSECS"]
        self.singular = "millisecond"
        self.plural = "milliseconds"

class Second(object):
    def __init__(self):
        self.scale = 1
        self.terms = ["s", "second", "Second", "SECOND", "seconds", "Seconds", "SECONDS", "sc", "Sc", "SC", "scs", "Scs", "SCS", "sec", "Sec", "SEC", "secs", "Secs", "SECS"]
        self.singular = "second"
        self.plural = "seconds"

class Minute(object):
    def __init__(self):
        self.scale = 60
        self.terms = ["m", "minute", "Minute", "MINUTE", "minutes", "Minutes", "MINUTES", "mn", "Mn", "MN", "mns", "Mns", "MNS", "min", "Min", "MIN", "mins", "Mins", "MINS"]
        self.singular = "minute"
        self.plural = "minutes"

class Hour(object):
    def __init__(self):
        self.scale = 3600
        self.terms = ["h", "hour", "Hour", "HOUR", "hours", "Hours", "HOURS", "hr", "Hr", "HR", "hrs", "Hrs", "HRS"]
        self.singular = "hour"
        self.plural = "hours"

class Day(object):
    def __init__(self):
        self.scale = 86400
        self.terms = ["d", "day", "Day", "DAY", "days", "Days", "DAYS", "dy", "Dy", "DY", "dys", "Dys", "DYS"]
        self.singular = "day"
        self.plural = "days"

class Week(object):
    def __init__(self):
        self.scale = 604800
        self.terms = ["w", "week", "Week", "WEEK", "weeks", "Weeks", "WEEKS", "wk", "Wk", "WK", "wks", "Wks", "WKS"]
        self.singular = "week"
        self.plural = "weeks"

class Month(object):
    def __init__(self):
        self.scale = 2635200
        self.terms = ["M", "month", "Month", "MONTH", "months", "Months", "MONTHS", "mth", "Mth", "MTH", "mths", "Mths", "MTHS", "mnth", "Mnth", "MNTH", "mnths", "Mnths", "MNTHS"]
        self.singular = "month"
        self.plural = "months"

class Year(object):
    def __init__(self):
        self.scale = 31536000
        self.terms = ["y", "year", "Year", "YEAR", "years", "Years", "YEARS", "yr", "Yr", "YR", "yrs", "Yrs", "YRS"]
        self.singular = "year"
        self.plural = "years"

class Decade(object):
    def __init__(self):
        self.scale = 315360000
        self.terms = ["D", "decade", "Decade", "DECADE", "decades", "Decades", "DECADES", "dc", "Dc", "DC", "dcs", "Decs", "DECS", "dec", "Dec", "DEC", "decs", "Decs", "DECS"]
        self.singular = "decade"
        self.plural = "decades"

class Century(object):
    def __init__(self):
        self.scale = 3153600000
        self.terms = ["c", "century", "Century", "CENTURY", "centuries", "Centuries", "CENTURIES", "cn", "Cn", "CN", "cns", "Cns", "CNS", "ct", "Ct", "CT", "cts", "Cts", "CTS", "cnt", "Cnt", "CNT", "cnts", "Cnts", "CNTS", "cent", "Cent", "CENT", "cents", "Cents", "CENTS"]
        self.singular = "century"
        self.plural = "centuries"


from timelength.scales import *

SEPERATORS = [
    ",",
    " ",
    "\t"
]

CONNECTORS = [
    "and"
]

SCALES = [Millisecond(), Second(), Minute(), Hour(), Day(), Week(), Month(), Year(), Decade(), Century()]

ABBREVIATIONS = [item for sublist in SCALES for item in sublist.terms]


class InvalidValue(Exception):
    '''Exception for when a value is not parseable.'''
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

class InvalidOrder(Exception):
    '''Exception for when Values and Scales are listed in an incorrect order.'''
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


from timelength.exceptions import *
from timelength.constants import *
from timelength.helper import *
from timelength.scales import *

class ParsedValues(object):
    def __init__(self, valid_values: list):
        self.valid = valid_values
        self.milliseconds = None
        self.seconds = None
        self.minutes = None
        self.hours = None
        self.days = None
        self.weeks = None
        self.months = None
        self.years = None
        self.decades = None
        self.centuries = None
        for length in valid_values:
            if length[1] in Millisecond().terms: 
                self.milliseconds = length[0]
            if length[1] in Second().terms: 
                self.seconds = length[0]
            if length[1] in Minute().terms: 
                self.minutes = length[0]
            if length[1] in Hour().terms: 
                self.hours = length[0]
            if length[1] in Day().terms: 
                self.days = length[0]
            if length[1] in Week().terms: 
                self.weeks = length[0]
            if length[1] in Month().terms: 
                self.months = length[0]
            if length[1] in Year().terms: 
                self.years = length[0]
            if length[1] in Decade().terms: 
                self.decades = length[0]
            if length[1] in Century().terms: 
                self.centuries = length[0]

    def __str__(self):
        return f"<ParsedValues {self.__dict__}>"

    def __repr__(self):
        return self.__str__()

class ParsedTimeLength(object):
    def __init__(self, total_seconds, valid_values):
        self.total_seconds = total_seconds
        self.valid_values = valid_values

class TimeLength(object):
    def __init__(self, passed_value, strict = False):
        self.passed_value = passed_value
        self.strict = strict
        self.__parsed_time_length = self.__parse(self.passed_value, self.strict)
        self.__valid_values = self.__parsed_time_length.valid_values
        self.total_seconds = self.__parsed_time_length.total_seconds
        self.parsed_value = ParsedValues(self.__valid_values)

    def __str__(self):
        return f"<TimeLength \"{self.passed_value}\">"

    def __repr__(self):
        return self.__str__()

    def __parse(self, passed_value, strict):
        buffer = ""
        potential_values = []
        valid_values = []
        skip_iteration = 0
        total_seconds = 0
        last_alphanum = None

        for index, char in enumerate(passed_value):
            # Skip iterations due to decimal or non-alphanumeric encounter
            if skip_iteration > 0:
                skip_iteration -= 1
                continue

            # Check if decimal number is encountered
            if (index + 2) < len(passed_value) and passed_value[index + 1] == "." and passed_value[index + 2].isdigit():
                if isfloat(char):
                    char += "."
                else:
                    char = "0."
                future_index = index + 2
                skip_iteration = 1
                while future_index < len(passed_value) and passed_value[future_index].isdigit():
                    char += passed_value[future_index]
                    future_index += 1
                    skip_iteration += 1
            
            # Catch special characters such as !#$+
            alphanum = check_alphanumeric(char)
            if char in SEPERATORS:
                if buffer and not buffer.strip() in CONNECTORS:
                    potential_values.append(buffer)
                buffer = ""
                last_alphanum = None
            elif not alphanum and (index + 1) < len(passed_value):
                if buffer:
                    potential_values.append(buffer)
                buffer = ""
                index += 1
                skip_iteration = 0
                to_check = passed_value[index]
                # Check if subsequent characters are also non-alphanumeric
                while index < len(passed_value) and to_check is not None and to_check not in SEPERATORS and not isfloat(to_check) and not check_alphanumeric(to_check):
                    char += passed_value[index]
                    skip_iteration += 1
                    index += 1
                    if index < len(passed_value):
                        to_check = passed_value[index]
                potential_values.append(char)
            # Only happens if the passed value ends in a single non-alphanumeric value
            elif not alphanum:
                if buffer:
                    potential_values.append(buffer)
                buffer = char
            else:
                if (alphanum and last_alphanum and alphanum != last_alphanum):
                    potential_values.append(buffer)
                    buffer = char
                else:
                    buffer += char
                last_alphanum = alphanum
        potential_values.append(buffer)

        index = 0
        input_length = None
        input_scale = None
        preceeding_num = False
        preceeding_alpha = True
        invalid_values = [item for item in potential_values if item and not isfloat(item) and item not in ABBREVIATIONS]
        potential_values = [item for item in potential_values if item and item not in invalid_values]
        # If invalid and strict, error
        # If not strict, ignore invalid and proceed
        if invalid_values and self.strict:
            raise InvalidValue(f"Input TimeLength \"{passed_value}\" contains {'invalid values' if len(invalid_values) > 1 else 'an invalid value'}: {invalid_values}")
        for item in potential_values:
            if isfloat(item):
                if preceeding_num is True and strict is True:
                    raise InvalidOrder(f"Input TimeLength \"{passed_value}\" contains multiple subsequent Values with no paired Scales: {potential_values}")
                input_length = float(item)
                preceeding_alpha = False
                preceeding_num = True
            else:
                if index == 0 and preceeding_alpha and strict is True:
                    raise InvalidOrder(f"Input TimeLength \"{passed_value}\" starts with a Scale rather than a Value: {potential_values}")
                elif preceeding_alpha is True and strict is True:
                    raise InvalidOrder(f"Input TimeLength \"{passed_value}\" contains multiple subsequent Scales with no paired Values: {potential_values}")
                if preceeding_num and item in ABBREVIATIONS:
                    # Calculate proper Value THEN Scale pairs
                    input_scale = item
                    for scale in SCALES:
                        if input_scale in scale.terms:
                            total_seconds += input_length * scale.scale
                            if input_length == 1:
                                input_scale = scale.singular
                            else:
                                input_scale = scale.plural
                            break
                    preceeding_num = False
                    preceeding_alpha = True
            index += 1
            # Append pairs to valid_values to be returned
            if input_length and input_scale:
                valid_values.append((input_length, input_scale))
                input_length = None
                input_scale = None
        
        if not valid_values:
            raise InvalidValue(f"Input TimeLength \"{passed_value}\" contains no valid Value and Scale pairs.")

        return ParsedTimeLength(total_seconds, valid_values)

    def to_milliseconds(self, max_precision = 2):
        return round(self.total_seconds / float(Millisecond().scale), max_precision)

    def to_seconds(self, max_precision = 2):
        return round(self.total_seconds, max_precision)

    def to_minutes(self, max_precision = 2):
        return round(self.total_seconds / float(Minute().scale), max_precision)

    def to_hours(self, max_precision = 2):
        return round(self.total_seconds / float(Hour().scale), max_precision)

    def to_days(self, max_precision = 2):
        return round(self.total_seconds / float(Day().scale), max_precision)

    def to_weeks(self, max_precision = 2):
        return round(self.total_seconds / float(Week().scale), max_precision)

    def to_months(self, max_precision = 2):
        return round(self.total_seconds / float(Month().scale), max_precision)

    def to_years(self, max_precision = 2):
        return round(self.total_seconds / float(Year().scale), max_precision)

    def to_decades(self, max_precision = 2):
        return round(self.total_seconds / float(Decade().scale), max_precision)

    def to_centuries(self, max_precision = 2):
        return round(self.total_seconds / float(Century().scale), max_precision)


from setuptools import setup, find_packages

def readme():
    with open("README.md") as f:
        return f.read()

setup(
    name = "timelength",
    version = "1.0.2",
    url = "https://github.com/Etorix0005/timelength",
    license = "MIT License",
    author = "Etorix",
    author_email = "admin@etorix.dev",
    description = "A Python package to parse human readable lengths of time.",
    long_description_content_type = "text/markdown",
    long_description = readme(),
    classifiers=["License :: OSI Approved :: MIT License",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.10"],

    packages = find_packages()
)